/*--------------------------------------------------------------------------
 * ordered heap / priority queue functions
 * $Id: heap.c 102 2006-07-08 09:06:45Z ncopa $
 * Copyright (c) 2005    Nathan Angelacos (nangel@users.sourceforge.net)
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 US
 * 
 ------------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "heap.h"

/* 
 * generic heap comparison function - this is used if no 
 * function is given
 */

int
_heapElementCompare (void *a, void *b)
{
  if (a > b)
    return (1);
  if (a < b)
    return (-1);
  return (0);
}

/* 
 * initialize a new heap stucture
 * Takes a pointer to the compare function
 * Returns a pointer to an initialized heap structure or NULL on failure
 */

heap_t *
heapInitialize (void *fn)
{
  heap_t *heap;

  heap = malloc (sizeof (heap_t));
  if (heap)
    {
      heap->capacity = 0;
      heap->size = 0;
      if (fn)
	{
	  heap->compare = fn;
	}
      else
	{
	  heap->compare = &_heapElementCompare;
	}
      heap->element = NULL;

    }
  return (heap);
}


/*
 * empty a heap and deallocate it
 */

void
heapFree (heap_t * h)
{
  void *ptr;
  
  if (h)
    {
      while ( (ptr = heapRemove(h)) ) {
	      free(ptr);
      	}
      if (h->element) free (h->element);
      free (h);
    }
}


/*
 * insert an element into a heap; allocating space for the data
 * returns a pointer to the element inserted, NULL on error
 */

void *
heapInsert (heap_t * h, void *data, size_t size )
{
  size_t pos;
  void *ptr;

  /* allocate space for the data */

  ptr = malloc(size);
  if ( !ptr ) {return NULL; }

  memcpy (ptr, data, size);
  
  h->size++;
  if (h->size >= h->capacity)
    {
      h->capacity += 1024;	/* allocate 1K of elements at a time */
      h->element = realloc (h->element, sizeof (void *) * h->capacity);
      if (h->element == NULL)
	{			/* malloc failed */
  	  free (ptr);
	  return (NULL);
	}
    }
  
  h->element[h->size] = ptr;

  /* siftup */
  pos = h->size;

  while ((pos > 1) && (h->compare (ptr, h->element[pos / 2]) <= 0))
    {
      h->element[pos] = h->element[pos / 2];
      pos = pos / 2;
    }
  h->element[pos] = ptr;

  return (ptr);
}


/*
 * delete the top most node from the heap
 * returns the element, or NULL if the heap is empty
 */

void *
heapRemove (heap_t * h)
{
  size_t pos, child;
  void *here, *top;

  if (h->size == 0)
    {
      return (NULL);
    }

  /*
  if ( h->size == 1)
	{ 
	h->size=0; 
	return (h->element[1]);
	}
  */
  top = h->element[1];
  
  h->element[1] = h->element[h->size];
  here = h->element[1];
	  
  /* siftdown */
  pos = 1;
  
  while (pos <= (h->size) / 2)
    {
      child = 2 * pos;
      if ((child < h->size)
	  && (h->compare (h->element[child], h->element[child + 1]) >= 0))
	{
	  child++;
	}
      if (h->compare (here, h->element[child]) < 0)
	{
	  break;
	}
      h->element[pos] = h->element[child];
      pos = child;
    }
  
  h->element[pos] = here;
  
  h->size--;
  return (top);
}


/* ------------------------------------------------------------------
 * This is the test framework - compile with gcc -o test -DTEST_FRAMEWORK
 */

#ifdef TEST_FRAMEWORK

#include <string.h>

int rcompare ( void *a, void *b ) {
	return ( strcmp ( a, b) * -1 );	
	}

main (void)
{
  heap_t *me;
  char *str;
  size_t count;
  char *alpha[] = { "Alpha", "Bravo", "Charlie", "Delta", "Echo", "Foxtrot", 
	  	"Golf", "Hotel", "India", "Juliet", "Kilo" };

  /* Initialize the heap, using the strcmp compare function */
  me = heapInitialize (&strcmp);

  heapInsert (me, "Oranges");
  heapInsert (me, "Fruit Loops");
  heapInsert (me, "Fruit Loops");
  heapInsert (me, "Crayons");
  heapInsert (me, "Doggies");
  heapInsert (me, "Balloons");
  heapInsert (me, "Eclairs");
  heapInsert (me, "French Fries");

  
  
  puts ("queue walking (should be in alphabetical order) ---");
  while (str = heapRemove (me))
    {
      puts (str);
    }
  heapFree (me);


  /* Initialze the heap, using the default compare function */
  me = heapInitialize (NULL);
  count = 11;
  while (count > 0) {
	  heapInsert (me, alpha[--count]);
  	}

  puts ("queue walking (should be in alphabetical order) ---");
  while ( str = heapRemove(me)) { puts (str); }
  heapFree(me);
  

  /* Initialze the heap, using a reverse compare function */
  me = heapInitialize ( &rcompare);
  count = 0;
  while (count < 11) {
	  heapInsertWithAlloc (me, alpha[count], strlen (alpha[count])+1);
	  count++;
  	}
	
  puts ("queue walking (should be in reverse alphabetical order) ---");
  while ( str = heapRemove(me)) { puts (str); free(str); }
  heapFree(me);
  


 me = heapInitialize ( &strcmp );
  heapInsertWithAlloc (me, "Eclairs", 8);
 
  str = heapRemove(me); free (str);
  str = heapRemove(me);
  str = heapRemove(me);
  str = heapRemove(me);
  str = heapRemove(me);
  heapFree(me);
  
}

#endif


