/*
 * --------------------------------------------------------------------------
 * sfic - simple file integrity checker 
 * $Id: sfic.c 110 2006-07-27 14:39:53Z nangel $ 
 * Copyright (c) 2005    Nathan Angelacos (nangel@users.sourceforge.net)
 *                       Natanel Copa (ncopa@users.sourceforge.net)
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * -------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <string.h>
#include <dirent.h>
#include <stdarg.h>

#include <tdb.h>

#include "heap.h"
#include "sha256.h"
#include "sfic.h"


const char *REVTAG = "$Id: sfic.c 110 2006-07-27 14:39:53Z nangel $";


enum sfic_operations
{ CMD_NULL, CMD_ADD, CMD_DELETE, CMD_LIST, CMD_COMPARE, CMD_FSCOMPARE,
  CMD_HELP
};
enum sfic_verbosity
{ NONE, TERSE, BRIEF, VERBOSE, DEBUG };


enum warn_types
{
  ONE_OP_ONLY_ERROR, NO_OP_ERROR, MALLOC_FAIL,
  FILE_OPEN_ERROR, FILE_MISSING, ILLEGAL_MASK_ERROR, IGNORING, ADDING,
  DELETING, NO_ERROR
};


/* Error messages */
char *sfic_error[] = {
  /* ONE_OP_ONLY_ERROR  */
  "Choose only one operation (Add, Delete, List, Compare)",
  /* NO_OP_ERROR        */ "Nothing to do",
  /* MALLOC_FAIL        */ "Out of memory",
  /* FILE_OPEN_ERROR    */ "Unable to open %s",
  /* FILE_MISSING       */ "Could not stat %s",
  /* ILLEGAL_MASK_ERROR */ "'%c' is not a valid identifer",
  /* IGNORING           */ "Ignoring: %s",
  /* ADDING             */ "Adding: %s",
  /* DELETING           */ "Deleting: %s",
  /* NO_ERROR           */ ""
};


/* Headings */
const char *sfic_entity[] = {
  "TYPE", "PERM", "INODE", "NLINK", "UID", "GID", "SIZE", "ATIME", "MTIME",
  "CTIME", "HASH", "NAME"
};


/* the string representation of the mask */
const char *sfic_maskstr = "tpinugsamchN";

/* String reprenstation of the file types (from ls) */
  /* FIFO, chardev, dir, blockdev, file, symlink, socket */
  /* man lstat for more info */
const char *sfic_dirtype = "pcdbfls";

/*
 * Command line directives When adding a long option, make sure
 * to update the short_options as well
 */

struct option ga_long_options[] = {
  {"add", required_argument, 0, 'a'},
  {"delete", required_argument, 0, 'd'},
  {"list", required_argument, 0, 'l'},
  {"new", required_argument, 0, 'n'},
  {"old", required_argument, 0, 'o'},
  {"quiet", no_argument, 0, 'q'},
  {"verbose", no_argument, 0, 'v'},
  {"file", required_argument, 0, 'f'},
  {"fast", no_argument, 0, 'F'},
  {"mask", required_argument, 0, 'm'},
  {"help", no_argument, 0, 'h'},
  {"recursive", no_argument, 0, 'R'},
  {"one-file-system", no_argument, 0, 'x'},
  {"ignore-trivial", no_argument, 0, 't'},
  {"ignore", required_argument, 0, 'i'},
  {0, 0, 0, 0}
};

const char *gs_short_options = "a:d:l:n:o:qvf:Fm:hRxti:";

/*
 * abort the program
 */
void
die_with_error (const char *s, ...)
{
  va_list p;
  va_start (p, s);
  vfprintf (stderr, s, p);
  va_end (p);
  fprintf (stderr, "\n");
  exit (-1);
}


/*
 * allocate memory or die, busybox style.
 */
void *
xmalloc (size_t size)
{
  void *buf;
  if ((buf = malloc (size)) == NULL)
    {
      die_with_error (sfic_error[MALLOC_FAIL]);
    }
  return buf;
}

/*
 * realloc memory, or die xmalloc style.
 */
void *
xrealloc (void *buf, size_t size)
{
  if ((buf = realloc (buf, size)) == NULL)
    {
      die_with_error (sfic_error[MALLOC_FAIL]);
    }
  return buf;
}

/* 
 * print an error message if the verbosity is >= the level specified.
 * If the string is NULL, verbosity is set to the level specified.
 *
 */
void
xreport (int level, const char *s, ...)
{
  static int verbosity = BRIEF;
  va_list p;

  if ( s == NULL ) {
	verbosity = level;
	return;
	}

  if ( level <= verbosity ) { 
  va_start (p, s);
  vfprintf (stderr, s, p);
  va_end (p);
  fprintf (stderr, "\n");
	}
}

/* 
 * closes a tdb file if open 
 */
void
closeTDB (TDB_CONTEXT * tdb)
{
  if (tdb)
    {
      tdb_close (tdb);
    }
}

/*
 * opens a tdb  - returns the tdb handle (context)
 */
TDB_CONTEXT *
openTDB (char *filename, int mode)
{
  TDB_CONTEXT *tdb = NULL;

  if ((filename) && (strlen (filename)))
    {
      tdb = tdb_open (filename, 0, 0, mode, 0600);
    }

  if (tdb == NULL)
    {
      die_with_error (sfic_error[FILE_OPEN_ERROR], filename);
    }

  return (tdb);
}

/*
 * return file type given a st_mode
 */
char
charFromMode (mode_t mode)
{
  return (sfic_dirtype[(mode >> 12) / 2]);
}


/*
 * Format a stat entry as a string.  If data is NULL, 
 * print headings instead. It is assumed the target string is
 * large enough to accept the string
 */
void
statToString (char *string, char *maskstr, file_info_t * data, char *filename)
{

  int i, j, count;
  static char buf[PATH_MAX];
  *string = '\0';

  for (count = 0; count < strlen (maskstr); count++)
    {
      if (count)
	{
	  sprintf (string, "%s", "\t");
	  string += strlen (string);
	}

      /* NULL data string makes us print a header */
      if (data == NULL)
	{
	  i = strchr (sfic_maskstr, (int) maskstr[count]) - sfic_maskstr;
	  if (i > -1)
	    {
	      sprintf (string, "%s", sfic_entity[i]);
	      string += strlen (string);
	    }
	}
      else
	{
	  /* Else, print the stats */

	  switch (maskstr[count])
	    {
	    case 't':
	      sprintf (string, "%c", charFromMode (data->statbuf.st_mode));
	      break;
	    case 'p':
	      sprintf (string, "%04o", data->statbuf.st_mode & 00007777);
	      break;
	    case 'i':
	      sprintf (string, "%4lu", data->statbuf.st_ino);
	      break;
	    case 'n':
	      sprintf (string, "%3u", (unsigned int) data->statbuf.st_nlink);
	      break;
	    case 'u':
	      sprintf (string, "%5u", data->statbuf.st_uid);
	      break;
	    case 'g':
	      sprintf (string, "%5u", data->statbuf.st_gid);
	      break;
	    case 's':
	      sprintf (string, "%8lu", data->statbuf.st_size);
	      break;
	    case 'a':
	      sprintf (string, "%8lu", data->statbuf.st_atime);
	      break;
	    case 'm':
	      sprintf (string, "%8lu", data->statbuf.st_mtime);
	      break;
	    case 'c':
	      sprintf (string, "%8lu", data->statbuf.st_ctime);
	      break;
	    case 'h':
	      memset (buf, 0, 1);
	      j = data->hashbits / 8;
	      for (i = 0; i < j; i++)
		{
		  sprintf (buf + (i * 2), "%02x", data->hash[i]);
		}
	      strcat (string, buf);
	      break;
	    case 'N':
	      sprintf (string, "%s", filename);
	      break;
	    }
	  string += strlen (string);
	}
    }
}



/*
 * formats a outputList format (what, null, pathspec) - 
 * internal function used by compareTdbContents
 */

void
formatOutputListEntry (char *buf, size_t size, int verbosity, char *str1, char *str2,
		       char *str3, char *name)
{

  switch (verbosity)
    {
    case TERSE:
      strncpy (buf, str1, size);
      break;
    case BRIEF:
      strncpy (buf, str2, size);
      break;
    default:
      strncpy (buf, str3, size);
      break;
    }
  strncat (buf, "\t", size);
  strncat (buf, name, size);
}


/*
 * given a mask, format maskstr to be a fixed-len printable version; 
 */
void
convertMaskToString (int mask, char *maskstr)
{
  int count;

  memset (maskstr, 0, strlen (sfic_maskstr) + 1);

  for (count = 0; count < strlen (sfic_maskstr); count++)
    {
      maskstr[count] = (mask & (1 << count)) ? sfic_maskstr[count] : ' ';
    }
}

/*
 * given a mask as a string, return int value of the mask 
 */
int
convertMaskFromString (char *str)
{
  int mask = 0;
  int count = 0;
  char *offset;

  while (str[count] != '\0')
    {
      offset = strchr (sfic_maskstr, (int) str[count]);
      if (offset == NULL)
	{
	  die_with_error (sfic_error[ILLEGAL_MASK_ERROR], str[count]);
	}
      mask += 1 << ((int) (offset - sfic_maskstr));
      count++;
    }

  return (mask);
}


/*
 * given a string, return value of the ignore mask
 */

unsigned
convert_ignore_from_string (char *str)
{
  int count = 0;
  unsigned mask = 0;

  while (str[count] != '\0')
    {
      switch (str[count])
	{
	case 'p':
	  mask = mask | S_IFIFO;
	  break;
	case 'c':
	  mask = mask | S_IFCHR;
	  break;
	case 'd':
	  mask = mask | S_IFDIR;
	  break;
	case 'b':
	  mask = mask | S_IFBLK;
	  break;
	case 'f':
	  mask = mask | S_IFREG;
	  break;
	case 'l':
	  mask = mask | S_IFLNK;
	  break;
	case 's':
	  mask = mask | S_IFSOCK;
	  break;
	default:
	  xreport (NONE, sfic_error[ILLEGAL_MASK_ERROR], str[count]);
	  break;

	}
      count++;
    }
  return (~mask) & S_IFMT;
}


/*
 * Compares two info types, and returns a bitmask of 
 * what does not match - a 0 indicates everything matches
 * only checks what is specified in checkmask
 */
unsigned int
lstatDelta (file_info_t * a, file_info_t * b, unsigned int checkmask)
{
  unsigned int retmask = 0;

  /* t */
  if (checkmask & STAT_TYPE)
    {
      retmask +=
	((a->statbuf.st_mode & 077770000) ==
	 (b->statbuf.st_mode & 077770000)) ? 0 : STAT_PERM;
    }
  /* p */
  if (checkmask & STAT_PERM)
    {
      retmask +=
	((a->statbuf.st_mode & 00007777) ==
	 (b->statbuf.st_mode & 00007777)) ? 0 : STAT_PERM;
    }

  /* i */
  if (checkmask & STAT_INODE)
    {
      retmask += (a->statbuf.st_ino == b->statbuf.st_ino ? 0 : STAT_INODE);
    }

  /* n */
  if (checkmask & STAT_NLINK)
    {
      retmask +=
	(a->statbuf.st_nlink == b->statbuf.st_nlink ? 0 : STAT_NLINK);
    }

  /* u */
  if (checkmask & STAT_UID)
    {
      retmask += (a->statbuf.st_uid == b->statbuf.st_uid ? 0 : STAT_UID);
    }

  /* g */
  if (checkmask & STAT_GID)
    {
      retmask += (a->statbuf.st_gid == b->statbuf.st_gid ? 0 : STAT_GID);
    }

  /* s */
  if (checkmask & STAT_SIZE)
    {
      retmask += (a->statbuf.st_size == b->statbuf.st_size ? 0 : STAT_SIZE);
    }

  /* a */
  if (checkmask & STAT_ATIME)
    {
      retmask +=
	(a->statbuf.st_atime == b->statbuf.st_atime ? 0 : STAT_ATIME);
    }

  /* m */
  if (checkmask & STAT_MTIME)
    {
      retmask +=
	(a->statbuf.st_mtime == b->statbuf.st_mtime ? 0 : STAT_MTIME);
    }

  /* c */
  if (checkmask & STAT_CTIME)
    {
      retmask +=
	(a->statbuf.st_ctime == b->statbuf.st_ctime ? 0 : STAT_CTIME);
    }

  /* h */
  if (checkmask & STAT_HASH)
    {
      /* SPECIAL - A NULL HASH ALWAYS MATCHES */
      if (a->hashbits && b->hashbits)
	{
	  retmask +=
	    ((memcmp (a->hash, b->hash, a->hashbits / 8) ==
	      0) ? 0 : STAT_HASH);
	}
    }
  return (retmask);

}


int
sha256FileByName (char *filename, unsigned char *sum)
{

  sha256_context ctx;
  int fp;
  int len;
  unsigned char *buf;


  memset (sum, 0, 32);

  if ((fp = open (filename, O_RDONLY)) != -1)
    {
      buf = xmalloc (65536);
      sha256_starts (&ctx);
      while ((len = read (fp, buf, 65536)) > 0)
	{
	  sha256_update (&ctx, buf, len);
	}
      close (fp);
      free (buf);
      sha256_finish (&ctx, sum);
      return (0);
    }
  return (-1);
}


void
printOutputListHeap (heap_t * list)
{
  char *str;

  while ((str = heapRemove (list)) != NULL)
    {
      printf ("%s\n", str + strlen (str) + 1);
      free (str);
    }
}



/*
 * Add a key (string) and data element to a heap - exit on error
 * SPECIAL: This will allocate a memory buffer the first time it is
 * called.  If called with a NULL heap, then the buffer is freed.
 * 
 */
void
addHeap (char *key, void *data, size_t size, heap_t * heap)
{
  static char *buf = NULL;
  static size_t buflen = 0;

/* a NULL heap frees buf mem */
  if (heap == NULL)
    {
      if (buf)
	free (buf);
      buf = NULL;
      buflen = 0;
      return;
    }

/* if buflen is too small, reallocate buffer */
  if ((strlen (key) + size + 2) > buflen)
    {
      buflen = buflen + ((strlen (key) + size + 2) * 1024) / 1024 + 1024;
      buf = xrealloc (buf, buflen);
    }

/* copy the key, then data */
  memcpy (buf, key, strlen (key) + 1);
  memcpy (buf + strlen (key) + 1, data, size);

/* and store */
  if (heapInsert (heap, buf, strlen (key) + 1 + size) == NULL)
    {
      die_with_error (sfic_error[MALLOC_FAIL]);
    }

}



/*
 * Add a directory's entries to heap, die on error
 */
void
addDirToHeap (char *dirspec, dev_t parentdev, unsigned flags,
	      heap_t * fileList)
{
  static char buf[PATH_MAX];
  DIR *dir;
  struct dirent *dp;
  struct stat statbuf;

  dir = opendir (dirspec);
  if (dir)
    {
      while ((dp = readdir (dir)))
	{
	  /* skip . and .. */
	  if ((!strcmp (dp->d_name, ".")) || (!strcmp (dp->d_name, "..")))
	    continue;

	  if (strcmp (dirspec, "/"))
	    {
	      snprintf (buf, sizeof (buf), "%s/%s", dirspec, dp->d_name);
	    }
	  else
	    {
	      snprintf (buf, sizeof (buf), "/%s", dp->d_name);
	    }

	  /* skip stuff not in the parent device if in one-file-system mode */
	  lstat (buf, &statbuf);
	  if ((statbuf.st_dev != parentdev) && (flags & FSPEC_NOMOUNTPOINTS))
	    continue;

	  addHeap (buf, NULL, 0, fileList);

	}

      closedir (dir);
    }

}

/*
 * given a file name, open the file and add to heap
 * if file name is "-", then read from stdin
 */
void
addHeapFromFile (char *source, heap_t * fileList)
{
  FILE *fp;
  static char str[PATH_MAX];
  char *ch;

  if (strcmp (source, "-") == 0)
    {
      fp = stdin;
    }
  else
    {
      fp = fopen (source, "r");
    }

  if (fp == NULL)
    {
      die_with_error (sfic_error[FILE_OPEN_ERROR], source);
    }

  while (fgets (str, 2047, fp))
    {
      ch = strchr (str, '\n');
      if (ch)
	{
	  *ch = '\0';
	}
      addHeap (str, NULL, 0, fileList);
    }

}

/*
 * Pull the next entry from the heap, and populate the info_type structure
 * If the entry is a dir and we are recursing, it may ADD entries to the heap
 * If the sfic flags indicate we are ignoring a type, it skips it and returns
 * the next valid item.  
 * Returns a pointer to the name, NULL if no names left.   
 * The caller must free the pointer when done.
 */
char *
get_next_file_spec (file_info_t * info, sfic_t * sf)
{
  char *str;
  int i;
  dev_t parentdev;

  memset (info, 0, sizeof (file_info_t));

  str = heapRemove (sf->fileList);
  if (str == NULL)
    return (NULL);

  /* if it contains a trailing /, strip it, except
   * if it is a lone / */
  i = strlen (str) - 1;
  if ((i > 0) && (str[i] == '/'))
    {
      str[i] = '\0';
    }


  if (lstat (str, &info->statbuf) == -1)
    {
      /* files missing, so get the next entry */
  xreport (BRIEF, sfic_error[FILE_MISSING], str);
      return (get_next_file_spec (info, sf));
    }

  /* If we are recursing, add the directory entries to the heap */
  if ((sf->flags & FSPEC_RECURSE) && (info->statbuf.st_mode & S_IFDIR))
    {
      parentdev = info->statbuf.st_dev;
      addDirToHeap (str, parentdev, sf->flags, sf->fileList);
    }

  /* are we ignoring this file type?  */
  if ((info->statbuf.st_mode & sf->ignore) == 0)
    {
	  xreport (VERBOSE, sfic_error[IGNORING], str);
      return (get_next_file_spec (info, sf));
    }

  /* regular file? try to set the hash */
  if ((sf->flags & FSPEC_DOHASH) && (info->statbuf.st_mode & S_IFREG)
      && (~info->statbuf.st_mode & S_IFLNK))
    {
      if (sha256FileByName (str, info->hash) == 0)
	{
	  info->hashbits = 256;
	}
    }

  return (str);
}

/* 
 * Add files to a previously opened tdb
 */
int
addHeapEntriesToDB (sfic_t * sf)
{
  file_info_t buf;
  TDB_DATA key, record;
  char *node;

  /* The data container doesn't change during the run of this function,
     just the contents.  Set once and forget. */
  record.dptr = (void *) &buf;
  record.dsize = sizeof (file_info_t);

  while ((node = get_next_file_spec (&buf, sf)))
    {
      key.dptr = node;
      key.dsize = strlen (key.dptr) + 1;

	  xreport (BRIEF, sfic_error[ADDING], key.dptr);
      tdb_store (sf->writetdb, key, record, TDB_REPLACE);
      free (node);
    }
  return (0);
}


/* 
 * Delete files from a previously opened tdb
 */

int
remove_heap_entries_from_db (sfic_t * sf)
{
  TDB_DATA key, record;
  file_info_t *info;
  char *tmp_ptr;
  char *name;
  int i;

  while ((name = heapRemove (sf->fileList)) != NULL)
    {
      key.dptr = name;
      key.dsize = strlen (key.dptr) + 1;

      record = tdb_fetch (sf->writetdb, key);
      info = (file_info_t *) record.dptr;
      /* are we ignoring this file type?  */
      if (info)
	{
	  if ((info->statbuf.st_mode & sf->ignore) == 0)
	    {
		  xreport (VERBOSE, sfic_error[IGNORING], key.dptr);
	    }
	  else
	    {
		  xreport (BRIEF, sfic_error[DELETING], key.dptr);
	      tdb_delete (sf->writetdb, key);
	    }
	  free (record.dptr);
	}

      /* If we are recursing, we delete children too.  Unfortunately, children
         can exist even if the parent doesn't, so we have to check for children
         on each name, not just on a successful delete */
      if (sf->flags & FSPEC_RECURSE)
	{

	  /* if it contains a trailing /, strip it */
	  i = strlen (name);
	  if ((i > 0) && (name[i - 1] == '/'))
	    {
	      i--;
	      name[i] = '\0';
	    }

	  key = tdb_firstkey (sf->writetdb);
	  while (key.dptr)
	    {
	      if (!strncmp (name, key.dptr, i))	/* found a match */
		{
		  if (key.dptr[i] == '/')	/* and the match has /<something> after it */
		    {

		      record = tdb_fetch (sf->writetdb, key);
		      info = (file_info_t *) record.dptr;
		      /* are we ignoring this file type?  */
		      if ((info->statbuf.st_mode & sf->ignore) == 0)
			{
			      xreport (VERBOSE, sfic_error[IGNORING], key.dptr);
			}
		      else
			{
			      xreport (BRIEF, sfic_error[DELETING], key.dptr);
			  tdb_delete (sf->writetdb, key);
			}
			free (record.dptr);
		    }
		}
	      tmp_ptr = key.dptr;
	      key = tdb_nextkey (sf->writetdb, key);
	      free (tmp_ptr);
	    }
	}			/* end recursive delete */
      free (name);
    }
  return (0);
}


/* 
 * Enumerate files in a previously opened tdb
 */
void
listEntriesFromDB (sfic_t * sf)
{
  TDB_DATA key, record;
  char *tmp_ptr;
  static char buf[PATH_MAX];		/* warning - No error checking done to verify the stats will be less than PATH_MAX */
  file_info_t *info;

  /* If verbose, print a header */
  if (sf->verbosity > BRIEF)
    {
      statToString (buf, sf->maskstr, (file_info_t *) NULL, NULL);
      addHeap ("", buf, strlen (buf) + 1, sf->outputList);
    }

  key = tdb_firstkey (sf->writetdb);
  while (key.dptr)
    {
      record = tdb_fetch (sf->writetdb, key);
      info = (file_info_t *) record.dptr;
      /* are we ignoring this file type?  */
      if ((info->statbuf.st_mode & sf->ignore) == 0)
	{
	      xreport (VERBOSE, sfic_error[IGNORING], key.dptr);
	}
      else
	{
	  statToString (buf, sf->maskstr, (file_info_t *) record.dptr,
			key.dptr);
	  addHeap (key.dptr, buf, strlen (buf) + 1, sf->outputList);
	}
      free (record.dptr);
      tmp_ptr = key.dptr;
      key = tdb_nextkey (sf->writetdb, key);
      free (tmp_ptr);
    }
}

/*
 * compares the readdtb with the files listed in the heap
 */
int
compare_live_filesystem (sfic_t *sf )
{
file_info_t info;
TDB_DATA key, record;
char *node;
static char buf[PATH_MAX + 64]; /* space for both file path and checksum */
unsigned mask;
int retval =0 ;
char maskstr[32];


while ((node = get_next_file_spec (&info, sf))) {
	/* are we ignoring this file type? */
	if ((info.statbuf.st_mode & sf->ignore ) == 0) {
		xreport (VERBOSE, sfic_error[IGNORING], node);
		continue;
		}

	key.dptr = node;
	key.dsize= strlen(node) + 1;
	record = tdb_fetch ( sf->readtdb, key);
 	if (record.dptr == NULL ) /* it doesn't exist in the db */
		{
	  	retval = -1;		/* we had a change */
	  	if (sf->verbosity > NONE)
	    	{
	        formatOutputListEntry (buf, sizeof(buf), sf->verbosity, "A", "Added ",
				     ">ADDED>>>>", node);
	        addHeap (node, buf, strlen (buf) + 1, sf->outputList);
	        }			/* end the verbosity is > NONE */
	}  
      else			/* not missing, so compare the two */
	{
          /* we got a  record, so we need to compare the two */
          mask =
            lstatDelta ((file_info_t *) record.dptr,
                        &info,
                        convertMaskFromString (sf->maskstr));
          /* If ignoring size/modtime changes on directoires and device nodes ... then ignore them */
          if ((sf->flags & FSPEC_NOTRIVIAL) &&
              (!strchr ("fl", charFromMode ((info.statbuf.st_mode)))))
            {
              mask =
                mask & ~(STAT_SIZE + STAT_ATIME + STAT_MTIME + STAT_CTIME);
            }


          if (mask)
            {
              retval = -1;      /* we had a change */
              if (sf->verbosity > NONE)
                {
                  convertMaskToString (mask, maskstr);
                  formatOutputListEntry (buf, sizeof(buf), sf->verbosity, "U", "Updated",
                                         maskstr, node);
                  addHeap (node, buf, strlen (buf) + 1,
                           sf->outputList);
                }               /* end if verbosity is NONE */
            }                   /* end if we have a mask */

	   free (record.dptr);
		
		}

	free (node);
	}
return retval;
}

/*
 * compares two databases, stuffs the output in the output queue
 */
int
compareTdbContents (sfic_t * sf)
{
  TDB_DATA new_key, new_record;
  TDB_DATA old_key, old_record;
  unsigned mask;
  char *tmp_ptr;
  char buf[1024];		/* Mask should never be > 1K */
  int retval = 0;
  char maskstr[32];
  /* For everything in old, check new */
  old_key = tdb_firstkey (sf->readtdb);
  while (old_key.dptr)
    {
      new_record = tdb_fetch (sf->writetdb, old_key);
      if (new_record.dptr == NULL)	/* missing - entry was deleted */
	{
	  retval = -1;		/* we had a change */
	  if (sf->verbosity > NONE)
	    {
	      formatOutputListEntry (buf, sizeof(buf), sf->verbosity, "D", "Deleted",
				     "<DELETED<<", old_key.dptr);
	      addHeap (old_key.dptr, buf, strlen (buf) + 1, sf->outputList);
	    }			/* end the verbosity is > NONE */
	}
      else			/* not missing, so compare the two */
	{
	  /* we got a new record, so we need to compare the two */
	  old_record = tdb_fetch (sf->readtdb, old_key);
	  mask =
	    lstatDelta ((file_info_t *) old_record.dptr,
			(file_info_t *) new_record.dptr,
			convertMaskFromString (sf->maskstr));
	  /* If ignoring size/modtime changes on directoires and device nodes ... then ignore them */
	  if ((sf->flags & FSPEC_NOTRIVIAL) &&
	      (!strchr
	       ("fl",
		charFromMode (((file_info_t *) new_record.dptr)->statbuf.
			      st_mode))))
	    {
	      mask =
		mask & ~(STAT_SIZE + STAT_ATIME + STAT_MTIME + STAT_CTIME);
	    }


	  if (mask)
	    {
	      retval = -1;	/* we had a change */
	      if (sf->verbosity > NONE)
		{
		  convertMaskToString (mask, maskstr);
		  formatOutputListEntry (buf, sizeof(buf), sf->verbosity, "U", "Updated",
					 maskstr, old_key.dptr);
		  addHeap (old_key.dptr, buf, strlen (buf) + 1,
			   sf->outputList);
		}		/* end if verbosity is NONE */
	    }			/* end if we have a mask */
	  free (old_record.dptr);
	  free (new_record.dptr);
	}			/* end if we have a new_record */
      tmp_ptr = old_key.dptr;
      old_key = tdb_nextkey (sf->readtdb, old_key);
      free (tmp_ptr);
    }				/* end main loop through each old filename */



  /* For everything in new, check if it exists in old */
  new_key = tdb_firstkey (sf->writetdb);
  while (new_key.dptr)
    {
      old_record = tdb_fetch (sf->readtdb, new_key);
      if (old_record.dptr == NULL)
	{
	  retval = -1;		/* we had a change */
	  if (sf->verbosity > NONE)
	    {
	      formatOutputListEntry (buf, sizeof(buf), sf->verbosity, "A", "Added ",
				     ">ADDED>>>>", new_key.dptr);
	      addHeap (new_key.dptr, buf, strlen (buf) + 1, sf->outputList);
	    }			/* end the file verbosity is > NONE */
	}			/* end if there was no old_record */
      else
	{
	  free (old_record.dptr);
	}
      tmp_ptr = new_key.dptr;
      new_key = tdb_nextkey (sf->writetdb, new_key);
      free (tmp_ptr);
    }				/* end the main loop */

  return (retval);
}

void
printUsage (sfic_t * globals)
{
  fprintf (stdout,
	   "\n%s - simple file integrity checker - version %s\n"
	   "For more information see: %s\n\n"
	   "Add records:\t\t--add (-a) <database> [ OPTIONS ] [ dirspecs | --file FILE ]\n"
	   "Delete records:\t\t--delete (-d) <database> [ OPTIONS ] [ dirspecs | --file FILE ]\n"
	   "List records:\t\t--list (-l) [ OPTIONS ] [--mask <mask>] <database>\n"
	   "Compare databases:\t--new (-n) <database> --old <database> [--mask <mask>] [-q | -v ]\n"
	   "\nOPTIONS:\n--file FILE is a file containing dirspecs, one per line, or \"-\" for stdin.\n"
	   "--quiet (-q) can be used more than once\n"
	   "--verbose (-v) can be used more than once\n"
	   "--fast (-F) prevents hashes from being generated on regular files\n"
	   "--resursive (-R) traverses directories\n"
	   "--one-file-system (-x) stops recursion on mount points\n"
	   "--ignore-trivial (-t) ignores size and time changes on directories\n"
	   "--ignore (-i) <types> ignore specified directory entries\n"
	   "\nmask specifies what to compare or list (default: %s)\n"
	   "\tN - file name\n\tt - file type\n\tp - permissions\n"
	   "\ti - inode\n\tn - num links\n\tu - user owner\n"
	   "\tg - group owner\n\ts - size\n"
	   "\ta - access time\n\tm - modified time\n"
	   "\tc - creation time\n\th - hash\n"
	   "\ntypes can be any combination of\n"
	   "\tb - blockdev\n\tc - chardev\n"
	   "\td - directory\n\tf - file\n"
	   "\tl - link\n\tp - FIFO (pipe)\n"
	   "\ts - socket\n\n", PACKAGE_NAME,
	   PACKAGE_VERSION, PACKAGE_BUGREPORT, globals->maskstr);
}

void
initializeSficStruct (sfic_t * sf)
{
  sf->writedbName = NULL;
  sf->readdbName = NULL;
  sf->fileList = heapInitialize (&strcmp);
  sf->outputList = heapInitialize (&strcmp);
  sf->operation = CMD_NULL;
  sf->verbosity = TERSE;
  sf->maskstr = "tpugsmhN";
  sf->separator = "\t";
  sf->flags = FSPEC_DOHASH;
  sf->writetdb = NULL;
  sf->readtdb = NULL;
  sf->ignore = convert_ignore_from_string ("");
}


int
parseCommandLine (int argc, char *argv[], sfic_t * globals)
{
  int c;
  int option_index = 0;
  while ((c = getopt_long (argc, argv, gs_short_options,
			   ga_long_options, &option_index)) != -1)
    {
      switch (c)
	{
	case 'a':
	case 'd':
	case 'l':
	case 'n':
	  globals->writedbName = optarg;
	  if (globals->operation != CMD_NULL)
	    {
	      die_with_error (sfic_error[ONE_OP_ONLY_ERROR]);
	    }
	  else
	    {
	      switch (c)
		{
		case 'a':
		  globals->operation = CMD_ADD;
		  break;
		case 'd':
		  globals->operation = CMD_DELETE;
		  break;
		case 'l':
		  globals->operation = CMD_LIST;
		  break;
		case 'n':
		  globals->operation = CMD_COMPARE;
		  break;
		}
	    }
	  break;
	case 'o':
	  globals->readdbName = optarg;
	  break;
	case 'q':
	  if (globals->verbosity > NONE)
	    {
	      globals->verbosity--;
	    }
	  break;
	case 'v':
	  if (globals->verbosity < DEBUG)
	    {
	      globals->verbosity++;
	    }
	  break;
	case 'm':
	  globals->maskstr = optarg;
	  break;
	case 'f':
	  addHeapFromFile (optarg, globals->fileList);
	  break;
	case 'F':
	  globals->flags = globals->flags & ~FSPEC_DOHASH;
	  break;
	case 'h':
	  globals->operation = CMD_HELP;
	  break;
	case 'R':
	  globals->flags = globals->flags | FSPEC_RECURSE;
	  break;
	case 'x':
	  globals->flags = globals->flags | FSPEC_NOMOUNTPOINTS;
	  break;
	case 't':
	  globals->flags = globals->flags | FSPEC_NOTRIVIAL;
	  break;
	case 'i':
	  globals->ignore = convert_ignore_from_string (optarg);
	  break;
	}
    }

   /* set the verbosity */
   xreport (globals->verbosity, NULL);

  /* if they gave a new db (for compare), but not an old db, then kill the cmd */
  if ((globals->operation == CMD_COMPARE) && (globals->readdbName == NULL))
    {
      globals->operation = CMD_NULL;
    }

  /* if they gave an old db but not a new db, then its a filesystem compare */
  if ((globals->operation == CMD_NULL) && (globals->readdbName))
    {
      globals->operation = CMD_FSCOMPARE;
    }

  /* whatever is left, add to file list (not file globs) */
  for (c = optind; c < argc; c++)
    {
      addHeap (argv[c], NULL, 0, globals->fileList);
    }

  return (optind);
}


int
main (int argc, char *argv[])
{
  static sfic_t globals;
  int retval = 0;
  initializeSficStruct (&globals);

  parseCommandLine (argc, argv, &globals);
  /* Which operation do we do */
  switch (globals.operation)
    {
    case CMD_HELP:
      printUsage (&globals);
      break;
    case CMD_NULL:
      die_with_error (sfic_error[NO_OP_ERROR]);
      break;
    case CMD_ADD:
      globals.writetdb = openTDB (globals.writedbName, O_RDWR + O_CREAT);
      addHeapEntriesToDB (&globals);
      break;
    case CMD_DELETE:
      globals.writetdb = openTDB (globals.writedbName, O_RDWR);
      remove_heap_entries_from_db (&globals);
      break;
    case CMD_LIST:
      globals.writetdb = openTDB (globals.writedbName, O_RDONLY);
      listEntriesFromDB (&globals);
      printOutputListHeap (globals.outputList);
      break;
    case CMD_COMPARE:
      globals.writetdb = openTDB (globals.writedbName, O_RDONLY);
      globals.readtdb = openTDB (globals.readdbName, O_RDONLY);
      retval = compareTdbContents (&globals);
      printOutputListHeap (globals.outputList);
      break;
    case CMD_FSCOMPARE:
      globals.readtdb = openTDB (globals.readdbName, O_RDONLY);
      retval = compare_live_filesystem(&globals);
      printOutputListHeap (globals.outputList);
      break;
    }


  closeTDB (globals.writetdb);
  closeTDB (globals.readtdb);
  heapFree (globals.fileList);
  heapFree (globals.outputList);
  addHeap (NULL, NULL, 0, NULL);
  return (retval);
}
